from .dataset_api import ACTION_FIXTURES as DATASET_API_ACTION_FIXTURES
from .filter_api import ACTION_FIXTURES as FILTER_API_ACTION_FIXTURES
from .layer_api import ACTION_FIXTURES as LAYER_API_ACTION_FIXTURES
from .map_api import ACTION_FIXTURES as MAP_API_ACTION_FIXTURES

SERIALIZED_ACTIONS = [
    *DATASET_API_ACTION_FIXTURES,
    *FILTER_API_ACTION_FIXTURES,
    *LAYER_API_ACTION_FIXTURES,
    *MAP_API_ACTION_FIXTURES,
]
